###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import sys
from collections import defaultdict, namedtuple

import dendropy

from gtdbtk.biolib_lite.newick import parse_label, create_label
from gtdbtk.biolib_lite.taxonomy import Taxonomy


class Decorate(object):
    """Decorate internal nodes with taxa labels."""

    def __init__(self):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')

        self.StatsTable = namedtuple('StatsTable',
                                     'node fmeasure precision recall taxa_in_lineage total_taxa num_leaves_with_taxa rogue_out rogue_in')

    def _fmeasure(self, tree, taxonomy):
        """Find node with highest F-measure for each taxon.
        
        Finds best placement for each taxon label
        by calculating the F-measure for every taxon
        at every node.
        
        Parameters
        ----------
        tree : Tree
          Dendropy Tree.
        taxonomy : d[extent_taxon_id] -> taxa list
          Taxon labels for extant taxa.
          
        Returns
        -------
        d[taxon] -> [(Node, F-measure, precision, recall_, ...]
            Node(s) with highest F-measure for each taxon.
        """

        # get named lineages/taxa at each taxonomic rank
        taxa_at_rank = Taxonomy().named_lineages_at_rank(taxonomy)

        # get extant taxa for each taxon label
        extent_taxa_with_label = {}
        for i, rank in enumerate(Taxonomy.rank_labels):
            extent_taxa_with_label[i] = Taxonomy().extant_taxa_for_rank(rank, taxonomy)

        # get parent taxon for each taxon:
        taxon_parents = Taxonomy().parents(taxonomy)

        # get number of leaves and taxon in each lineage
        self.logger.info('Calculating taxa within each lineage.')
        for node in tree.preorder_node_iter():
            num_leaves = 0
            taxa_count = defaultdict(lambda: defaultdict(int))
            for leaf in node.leaf_iter():
                num_leaves += 1
                for rank_index, taxon in enumerate(taxonomy[leaf.taxon.label]):
                    if taxon != Taxonomy.rank_prefixes[rank_index]:
                        taxa_count[rank_index][taxon] += 1

            node.num_leaves = num_leaves
            node.taxa_count = taxa_count

        taxa_in_tree = defaultdict(int)
        for leaf in tree.leaf_node_iter():
            for taxon in taxonomy[leaf.taxon.label]:
                taxa_in_tree[taxon] += 1

        # find node with best F-measure for each taxon
        fmeasure_for_taxa = {}
        for rank_index in range(0, len(Taxonomy.rank_labels)):
            # if rank_index == 6: #*** skip species
            #    continue 
            self.logger.info('Processing {:,} taxa at {} rank.'.format(
                len(taxa_at_rank[rank_index]),
                Taxonomy.rank_labels[rank_index].capitalize()))

            for taxon in taxa_at_rank[rank_index]:
                if rank_index == 0:
                    # processing taxa at the domain is a special case
                    taxon_parent_node = tree.seed_node
                else:
                    # find first named parent 
                    # e.g., Cyanobacteria for Synechococcales in d__Bacteria;p__Cyanobacteria;c__;o__Synechococcales
                    parent_taxon = 'x__'
                    parent_index = rank_index - 1
                    while len(parent_taxon) == 3 and parent_index != -1:
                        parent_taxon = taxon_parents[taxon][parent_index]
                        parent_index -= 1

                    if parent_taxon in fmeasure_for_taxa:
                        # only need to process the lineage below the parent node,
                        # but must take the MRCA if the placement of the parent
                        # taxon is unresolved
                        parent_nodes = []
                        for stat_table in fmeasure_for_taxa[parent_taxon]:
                            parent_nodes.append(stat_table.node)

                        if len(parent_nodes) == 1:
                            taxon_parent_node = parent_nodes[0]
                        else:
                            taxa = []
                            for p in parent_nodes:
                                taxa += [leaf.taxon for leaf in p.leaf_iter()]
                            taxon_parent_node = tree.mrca(taxa=taxa)

                        if taxon_parent_node.taxa_count[rank_index][taxon] < 0.5 * taxa_in_tree[taxon]:
                            # substantial portion of genomes for taxon fall outside 
                            # the parent lineages so best search the entire tree
                            taxon_parent_node = tree.seed_node
                    else:
                        # the parent for this taxon was not placed so
                        # it can be ignored (e.g., bacterial phylum in archaeal tree)
                        continue

                cur_taxon_fmeasure = -1
                cur_taxa = set(extent_taxa_with_label[rank_index][taxon])
                total_taxa = len(cur_taxa)

                for node in taxon_parent_node.preorder_iter():
                    taxa_in_lineage = node.taxa_count[rank_index][taxon]
                    num_leaves_with_taxa = sum(node.taxa_count[rank_index].values())

                    if taxa_in_lineage != 0 and num_leaves_with_taxa != 0:
                        precision = float(taxa_in_lineage) / num_leaves_with_taxa
                        recall = float(taxa_in_lineage) / total_taxa
                        fmeasure = (2 * precision * recall) / (precision + recall)

                        if fmeasure >= cur_taxon_fmeasure:
                            node_taxa = set([l.taxon.label for l in node.leaf_iter()])
                            rogue_out = cur_taxa - node_taxa
                            rogue_in = []
                            for gid in node_taxa - cur_taxa:
                                if taxonomy[gid][rank_index] != Taxonomy.rank_prefixes[rank_index]:
                                    rogue_in.append(gid)

                            stat_table = self.StatsTable(node=node,
                                                         fmeasure=fmeasure,
                                                         precision=precision,
                                                         recall=recall,
                                                         taxa_in_lineage=taxa_in_lineage,
                                                         total_taxa=total_taxa,
                                                         num_leaves_with_taxa=num_leaves_with_taxa,
                                                         rogue_out=rogue_out,
                                                         rogue_in=rogue_in)

                            if fmeasure > cur_taxon_fmeasure:
                                cur_taxon_fmeasure = fmeasure
                                fmeasure_for_taxa[taxon] = [stat_table]
                            elif fmeasure == cur_taxon_fmeasure:
                                fmeasure_for_taxa[taxon].append(stat_table)

        return fmeasure_for_taxa

    def _strip_taxon_labels(self, tree):
        """Remove any previous taxon labels.
        
        Parameters
        ----------
        tree : Tree
          Dendropy Tree.
        """

        for node in tree.internal_nodes():
            support, _taxon, _aux_info = parse_label(node.label)
            if support is not None:
                node.label = create_label(support, None, None)
            else:
                node.label = None

    def _assign_taxon_labels(self, fmeasure_for_taxa):
        """Assign taxon labels to nodes.
        
        Parameters
        ----------
        fmeasure_for_taxa : d[taxon] -> [(Node, F-measure, precision, recall), ...]
          Node with highest F-measure for each taxon.
          
        Returns
        -------
        set
            Taxon labels placed in tree.
        """

        placed_taxon = set()
        for taxon in Taxonomy().sort_taxa(fmeasure_for_taxa.keys()):
            if len(fmeasure_for_taxa[taxon]) == 1:
                placed_taxon.add(taxon)

                stat_table = fmeasure_for_taxa[taxon][0]
                node = stat_table.node
                fmeasure = stat_table.fmeasure
                precision = stat_table.precision
                recall = stat_table.recall

                support, taxon_label, aux_info = parse_label(node.label)
                if taxon_label:
                    taxon_label += '; ' + taxon
                else:
                    taxon_label = taxon
                node.label = create_label(support, taxon_label, aux_info)

        return placed_taxon

    def _write_statistics_table(self, fmeasure_for_taxa, taxonomy, out_table):
        """Write table containing statistics for each taxon.
        
        Parameters
        ----------
        fmeasure_for_taxa : d[taxon] -> [(Node, F-measure, precision, recall)]
          Node with highest F-measure for each taxon.
        taxonomy : d[unique_id] -> [d__<taxon>; ...; s__<taxon>]
          Taxonomic information for taxa in tree of interest.
        out_table : str
          Output table to write statistics for assigned labels.  
        """

        # get extent taxa
        extant_taxa = Taxonomy().extant_taxa(taxonomy)

        fout_table = open(out_table, 'w')
        fout_table.write('Taxon\tNo. Expected in Tree\tF-measure\tPrecision\tRecall')
        fout_table.write('\tNo. Genomes from Taxon\tNo. Genome In Lineage')
        fout_table.write('\tRogue out\tRogue in\n')
        for taxon in Taxonomy().sort_taxa(fmeasure_for_taxa.keys()):
            if len(fmeasure_for_taxa[taxon]) != 1:
                self.logger.error('Multiple positions specified for taxon label.')
                sys.exit()

            num_genomes = len(extant_taxa[taxon])

            stat_table = fmeasure_for_taxa[taxon][0]
            fout_table.write('%s\t%d\t%.4f\t%.4f\t%.4f\t%d\t%d\t%s\t%s\n' % (
                taxon,
                num_genomes,
                stat_table.fmeasure,
                stat_table.precision,
                stat_table.recall,
                stat_table.taxa_in_lineage,
                stat_table.num_leaves_with_taxa,
                ','.join(stat_table.rogue_out),
                ','.join(stat_table.rogue_in)))

        fout_table.close()

    def _leaf_taxa(self, leaf):
        """Get taxonomic information for leaf node.
        
        Parameters
        ----------
        leaf : Node
          Node in tree.
          
        Returns
        -------
        list
          Taxa for leaf in rank order.
        """

        leaf_taxa = []

        parent = leaf
        while parent:
            _support, taxon, _aux_info = parse_label(parent.label)

            if taxon:
                for t in taxon.split(';')[::-1]:
                    leaf_taxa.append(t.strip())

            parent = parent.parent_node

        ordered_taxa = leaf_taxa[::-1]

        # fill in missing ranks
        last_rank = ordered_taxa[-1][0:3]
        for i in range(Taxonomy.rank_prefixes.index(last_rank) + 1, len(Taxonomy.rank_prefixes)):
            ordered_taxa.append(Taxonomy.rank_prefixes[i])

        return ordered_taxa

    def _write_taxonomy(self, tree, out_taxonomy):
        """Write taxonomy decorated on tree to file.
        
        Parameters
        ----------
        tree : Tree
          Dendropy Tree.
        out_taxonomy : str
          Output file.
        """

        fout = open(out_taxonomy, 'w')
        for leaf in tree.leaf_node_iter():
            taxa = self._leaf_taxa(leaf)
            fout.write('%s\t%s\n' % (leaf.taxon.label, '; '.join(taxa)))

        fout.close()

    def run(self, input_tree, input_taxonomy, output_tree):
        """Decorate internal nodes with taxa labels.

        Parameters
        ----------
        input_tree : str
          Tree to decorate
        input_taxonomy : dict
          Dictionary indicating taxa for genomes.
        output_tree: str
          Name of output tree.
        """

        # read tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        # remove any previous taxon labels
        self.logger.info('Removing any previous internal node labels.')
        self._strip_taxon_labels(tree)

        # trim taxonomy to taxa in tree
        taxonomy = {}
        for leaf in tree.leaf_node_iter():
            taxonomy[leaf.taxon.label] = input_taxonomy.get(leaf.taxon.label, Taxonomy.rank_prefixes)

        # find best placement for each taxon based on the F-measure statistic
        self.logger.info('Calculating F-measure statistic for each taxa.')
        fmeasure_for_taxa = self._fmeasure(tree, taxonomy)

        # select most terminal placement for each taxon in order to be conservative
        ambiguous_placements = set()
        for taxon, fmeasures in fmeasure_for_taxa.items():
            if len(fmeasures) != 1:
                ambiguous_placements.add(taxon)
                fmeasure_for_taxa[taxon] = [fmeasures[-1]]

        if len(ambiguous_placements) > 0:
            self.logger.warning('There are {:,} taxa with multiple placements '
                                'of equal quality.'.format(len(ambiguous_placements)))
            self.logger.warning('These were resolved by placing the label at '
                                'the most terminal position.')
            self.logger.warning('Ideally, taxonomic assignment of all genomes '
                                'should be established before tree decoration.')

        # place all labels on tree
        self.logger.info('Placing labels on tree.')
        placed_taxon = self._assign_taxon_labels(fmeasure_for_taxa)

        # write statistics for placed taxon labels
        self.logger.info('Writing out statistics for taxa.')
        out_table = output_tree + '-table'
        self._write_statistics_table(fmeasure_for_taxa, taxonomy, out_table)

        # output taxonomy of extant taxa on tree
        self.logger.info('Writing out inferred taxonomy for each genome.')
        out_taxonomy = output_tree + '-taxonomy'
        self._write_taxonomy(tree, out_taxonomy)

        # output decorated tree
        self.logger.info('Writing out decorated tree.')
        tree.write_to_path(output_tree,
                           schema='newick',
                           suppress_rooting=True,
                           unquoted_underscores=True)



from gtdbtk.biolib_lite.common import canonical_gid
from gtdbtk.config.config import RADII_FILE


class GTDBRadiiFile(object):
    """A wrapper for the gtdb_radii.tsv file included in the reference data."""
    path = RADII_FILE

    def __init__(self):
        self._rep_idx = None
        self._species_idx = None
        self._read()

    def _read(self):
        """Read the file and create any data."""
        self._rep_idx, self._species_idx = dict(), dict()
        with open(self.path) as fh:
            for line in fh.readlines():
                species, genome, ani = line.strip().split('\t')
                genome = canonical_gid(genome)
                ani = float(ani)
                self._rep_idx[genome] = {'species': species, 'ani': ani}
                self._species_idx[species] = {'rep': genome, 'ani': ani}

    def get_species_ani(self, species):
        """Returns the ANI for a specific species.

        Parameters
        ----------
        species : str
            The name of the species.

        Returns
        -------
        float
            The ANI value for that species.
        """
        return self._species_idx[species]['ani']

    def get_species_rep(self, species):
        """Returns the representative genome for this species.

        Parameters
        ----------
        species : str
            The name of the species.

        Returns
        -------
        str
            The genome which represents this species.
        """
        return self._species_idx[species]['rep']

    def get_rep_ani(self, rep):
        """Returns the ANI for the species circumscribed by this rep.

        Parameters
        ----------
        rep : str
            The representative genome.

        Returns
        -------
        float
            The ANI value for that species.
        """
        return self._rep_idx[rep]['ani']

    def get_rep_species(self, rep):
        """Returns the species that this genome represents..

        Parameters
        ----------
        rep : str
            The representative genome.

        Returns
        -------
        str
            The species represented.
        """
        return self._rep_idx[rep]['species']



import os
from typing import Optional

from gtdbtk.config.output import TRANSLATION_TABLE_SUFFIX
from gtdbtk.exceptions import GTDBTkExit


class TlnTableFile(object):
    """The translation table summary file generated by each genome.."""

    def __init__(self, out_dir: str, gid: str,
                 best_tln_table: Optional[int] = None,
                 coding_density_4: Optional[float] = None,
                 coding_density_11: Optional[float] = None):
        self.path = self.get_path(out_dir, gid)
        self._best_tln_table = best_tln_table
        self._coding_density_4 = coding_density_4
        self._coding_density_11 = coding_density_11

    @property
    def best_tln_table(self):
        return self._best_tln_table

    @best_tln_table.setter
    def best_tln_table(self, v):
        try:
            self._best_tln_table = int(v)
        except ValueError:
            raise GTDBTkExit(f'Invalid translation table: {v} for {self.path}')

    @property
    def coding_density_4(self):
        return self._coding_density_4

    @coding_density_4.setter
    def coding_density_4(self, v):
        try:
            self._coding_density_4 = float(v)
        except ValueError:
            raise GTDBTkExit(f'Invalid coding density: {v} for {self.path}')

    @property
    def coding_density_11(self):
        return self._coding_density_11

    @coding_density_11.setter
    def coding_density_11(self, v):
        try:
            self._coding_density_11 = float(v)
        except ValueError:
            raise GTDBTkExit(f'Invalid coding density: {v} for {self.path}')

    @staticmethod
    def get_path(out_dir: str, gid: str):
        # return os.path.join(out_dir, f'{gid}{TRANSLATION_TABLE_SUFFIX}')
        # TODO: One day this should use the above, will probably break other workflows.
        return os.path.join(out_dir, f'prodigal{TRANSLATION_TABLE_SUFFIX}')

    def read(self):
        with open(self.path, 'r') as fh:
            for line in fh.readlines():
                idx, val = line.strip().split('\t')
                if idx == 'best_translation_table':
                    self.best_tln_table = val
                elif idx == 'coding_density_4':
                    self.coding_density_4 = val
                elif idx == 'coding_density_11':
                    self.coding_density_11 = val

    def write(self):
        with open(self.path, 'w') as fh:
            fh.write(f'best_translation_table\t{self.best_tln_table}\n')
            fh.write(f'coding_density_4\t{self.coding_density_4}\n')
            fh.write(f'coding_density_11\t{self.coding_density_11}\n')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import os
import shutil
import tempfile
import unittest

from gtdbtk.exceptions import *
from gtdbtk.markers import Markers


class TestMarkers(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.markers = Markers(cpus=10)

    def tearDown(self):
        shutil.rmtree(self.dir_tmp)
        del self.markers

    def test__apply_mask(self):
        """ Test that applying a mask to an MSA works as expected with no pruning """
        gtdb_msa = {
            'ref_1': 'DDRMIENV',
            'ref_2': 'RDRMIE-V',
        }
        user_msa = {
            'usr_1': '-DRMIENV',
        }
        min_perc_aa = 0.0
        msa_mask = os.path.join(self.dir_tmp, 'msa_mask.mask')
        with open(msa_mask, 'w') as fm:
            fm.write('11100011')

        output_seqs, pruned_seqs = self.markers._apply_mask(gtdb_msa, user_msa, msa_mask, min_perc_aa)

        expected_out = {'ref_1': 'DDRNV',
                        'ref_2': 'RDR-V',
                        'usr_1': '-DRNV'}

        self.assertDictEqual(output_seqs, expected_out)

    def test__apply_mask_prune(self):
        """ Test that applying a mask to an MSA works as expected with pruning """
        gtdb_msa = {
            'ref_1': 'DDRMIENV',
            'ref_2': 'RDRMIE-V',
        }
        user_msa = {
            'usr_1': '-DRMIENV',
            'usr_2': '---MIE-V'
        }
        min_perc_aa = 0.5
        msa_mask = os.path.join(self.dir_tmp, 'msa_mask.mask')
        with open(msa_mask, 'w') as fm:
            fm.write('11100011')

        output_seqs, pruned_seqs = self.markers._apply_mask(gtdb_msa, user_msa, msa_mask, min_perc_aa)

        expected_out = {'ref_1': 'DDRNV',
                        'ref_2': 'RDR-V',
                        'usr_1': '-DRNV'}
        expected_pruned = {'usr_2': '----V'}

        self.assertDictEqual(output_seqs, expected_out)
        self.assertDictEqual(pruned_seqs, expected_pruned)

    def test__apply_mask__mask_len_mismatch(self):
        """ Test that an exception is thrown when the mask length does not equal the MSA """
        gtdb_msa = {
            'ref_1': 'ADRMIE'
        }
        user_msa = {
            'usr_1': 'AD-MIE'
        }
        min_perc_aa = 0.1
        msa_mask = os.path.join(self.dir_tmp, 'msa_mask.mask')
        with open(msa_mask, 'w') as fm:
            fm.write('101')

        self.assertRaises(MSAMaskLengthMismatch, self.markers._apply_mask, gtdb_msa, user_msa, msa_mask, min_perc_aa)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest

import dendropy

import gtdbtk.config.config as Config
from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.classify import Classify
from gtdbtk.config.output import *
from gtdbtk.io.pplacer_classification import PplacerClassifyFileAR122


class TestClassify(unittest.TestCase):

    def setUp(self):
        self.classify = Classify()
        self.out_dir = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.prefix = 'gtdbtk'
        self.pplacer_dir_reference = 'tests/data/pplacer_dir_reference'
        self.aln_dir_ref = 'tests/data/align_dir_reference/align'
        self.user_msa_file = os.path.join(self.aln_dir_ref, 'gtdbtk.ar122.user_msa.fasta')
        self.taxonomy_file = Config.TAXONOMY_FILE
        self.gtdb_taxonomy = Taxonomy().read(self.taxonomy_file)

    def tearDown(self):
        shutil.rmtree(self.out_dir)

    def test_standardise_taxonomy(self):
        taxstring = 'p__phylum1;c_class1'
        marker_set = 'bac120'
        new_taxstring = self.classify.standardise_taxonomy(
            taxstring, marker_set)
        self.assertEqual(
            new_taxstring, 'd__Bacteria;p__phylum1;c_class1;o__;f__;g__;s__')

        # Test that the correct domain is returned.
        self.assertEqual(self.classify.standardise_taxonomy('p__P;c__C;o__O;f__F;g__G;s__S', 'bac120'),
                         'd__Bacteria;p__P;c__C;o__O;f__F;g__G;s__S')
        self.assertEqual(self.classify.standardise_taxonomy('p__P;c__C;o__O;f__F;g__G;s__S', 'ar122'),
                         'd__Archaea;p__P;c__C;o__O;f__F;g__G;s__S')

        # Remove ranks and check
        rank_order = {'p': 0, 'c': 1, 'o': 2, 'f': 3, 'g': 4, 's': 5}
        rank_lst = ['p__P', 'c__C', 'o__O', 'f__F', 'g__G', 's__S']
        ranks = {'p': 'P', 'c': 'C', 'o': 'O', 'f': 'F', 'g': 'G', 's': 'S'}
        dom_info = {'d__Bacteria': 'bac120', 'd__Archaea': 'ar122'}

        for k in range(1, len(ranks) - 1):
            for cur_domain in ('d__Bacteria', 'd__Archaea'):
                ranks_selected = rank_lst[0:-k]
                expected = list()
                test_lst = list()
                for cur_rank, _ in sorted(rank_order.items(), key=lambda x: [1]):
                    if cur_rank in ranks_selected:
                        test_lst.append(f'{cur_rank}__{ranks[cur_rank]}')
                        expected.append(f'{cur_rank}__{ranks[cur_rank]}')
                    else:
                        expected.append(f'{cur_rank}__')

                expected_str = f'{cur_domain};{";".join(expected)}'
                test_str = ";".join(test_lst)

                cur_dom = dom_info[cur_domain]
                test_value = self.classify.standardise_taxonomy(test_str, cur_dom)
                self.assertEqual(expected_str, test_value)

    def test_get_pplacer_taxonomy(self):
        if not os.path.exists(self.out_dir):
            os.makedirs(self.out_dir)
        tree = dendropy.Tree.get_from_path(os.path.join(os.getcwd(), self.pplacer_dir_reference,
                                                        'gtdbtk.ar122.classify.tree'),
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)
        pplacer_classify_file = PplacerClassifyFileAR122(self.out_dir, self.prefix)
        self.classify._get_pplacer_taxonomy(pplacer_classify_file, 'ar122', self.user_msa_file, tree)
        results = {}

        with open(os.path.join(self.out_dir, PATH_AR122_PPLACER_CLASS.format(prefix=self.prefix)), 'r') as f:
            for line in f:
                infos = line.strip().split('\t')
                results[infos[0]] = infos[1]
        self.assertTrue(len(results) == 3)
        self.assertTrue('genome_1' in results)
        self.assertTrue('genome_2' in results)
        self.assertTrue('genome_3' in results)
        self.assertEqual(results.get(
            'genome_1'), 'd__Archaea;p__Euryarchaeota;c__Methanobacteria;o__Methanobacteriales;f__Methanobacteriaceae;g__Methanobrevibacter;s__')

    def test_place_genomes(self):
        tree_file = self.classify.place_genomes(
            self.user_msa_file, 'ar122', self.out_dir, self.prefix)
        with open(tree_file, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertTrue(last_line.startswith('('))
        self.assertTrue(last_line.endswith('d__Archaea;'))

    def test_formatnote(self):
        first3genomes = list(self.gtdb_taxonomy.keys())[:3]
        sorted_dict = ((first3genomes[0], {'ani': 98.5, 'af': 1.0}), (first3genomes[1], {
            'ani': 92.6, 'af': 1.0}), (first3genomes[2], {'ani': 90.3, 'af': 1.3}))
        labels = [first3genomes[0]]
        note_list = self.classify._formatnote(sorted_dict, labels)
        self.assertTrue(first3genomes[1] in note_list[0])
        self.assertTrue(first3genomes[2] in note_list[1])
        self.assertTrue(note_list[0].endswith(', 92.6, 1.0'))
        self.assertTrue(note_list[1].endswith(', 90.3, 1.3'))

    def test_calculate_red_distances(self):
        tree = os.path.join(self.pplacer_dir_reference,
                            'gtdbtk.ar122.classify.tree')
        result_tree = self.classify._calculate_red_distances(
            tree, self.out_dir)
        egs2 = [eg.length for eg in result_tree.postorder_edge_iter()
                if eg.length is not None]

        self.assertTrue(sum(egs2) / len(egs2) < 0.1)


if __name__ == '__main__':
    unittest.main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import ntpath
import os
import re
import sys

from tqdm import tqdm

from gtdbtk.config.config import LOG_TASK
from .common import make_sure_path_exists


def supports_colour():
    """Check that the current terminal supports colour.

    Returns
    -------
    bool
        True if the terminal supports colour, False otherwise.

    References
    ----------
        https://github.com/django/django/blob/master/django/core/management/color.py

    """
    try:
        supported_platform = sys.platform != 'win32' or 'ANSICON' in os.environ
        is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
        return supported_platform and is_a_tty
    except Exception:
        return False


def colour(to_fmt, attr=None, fg=None, bg=None):
    """Format a string using ANSI colour encoding, if the terminal supports it.

    Parameters
    ----------
    to_fmt : str
        The string to be formatted.
    attr : list
        A list of attributes to be applied.
    fg : str
        The foreground colour to apply.
    bg : str
        The background colour to apply.

    Returns
    -------
    str
        A string formatted according to the specifications.
    """
    if not supports_colour():
        return to_fmt
    else:
        options_attr = {'reset': 0, 'bright': 1, 'dim': 2, 'underscore': 4,
                        'blink': 5, 'reverse': 7, 'hidden': 8}
        options_col = {'black': 0, 'red': 1, 'green': 2, 'yellow': 3, 'blue': 4,
                       'magenta': 5, 'cyan': 6, 'white': 7}
        options = list() if not attr else [str(options_attr[x]) for x in attr]
        if fg:
            options.append(str(30 + options_col[fg]))
        if bg:
            options.append(str(40 + options_col[bg]))
        return '\x1b[{}m{}\x1b[0m'.format(';'.join(options), to_fmt)


def logger_setup(log_dir, log_file, program_name, version, silent, debug=False):
    """Setup loggers.

    Two logger are setup which both print to the stdout and a
    log file when the log_dir is not None. The first logger is
    named 'timestamp' and provides a timestamp with each call,
    while the other is named 'no_timestamp' and does not prepend
    any information. The attribution 'is_silent' is also added
    to each logger to indicate if the silent flag is thrown.

    Parameters
    ----------
    log_dir : str
        Output directory for log file.
    log_file : str
        Desired name of log file.
    program_name : str
        Name of program.
    version : str
        Program version number.
    silent : boolean
        Flag indicating if output to stdout should be suppressed.
    """

    class SpecialFormatter(logging.Formatter):
        """Terminal output rules"""
        default_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                        format(colour('INFO:', ['bright'])),
                                        datefmt="%Y-%m-%d %H:%M:%S")
        debug_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                      format(colour('DEBUG:', ['bright'], 'green')),
                                      datefmt="%Y-%m-%d %H:%M:%S")
        info_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                     format(colour('INFO:', ['bright'])),
                                     datefmt="%Y-%m-%d %H:%M:%S")
        warn_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                     format(colour('WARNING:', ['bright'], 'yellow')),
                                     datefmt="%Y-%m-%d %H:%M:%S")
        err_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                    format(colour('ERROR:', ['bright'], 'red')),
                                    datefmt="%Y-%m-%d %H:%M:%S")
        task_fmt = logging.Formatter(fmt="[%(asctime)s] {} %(message)s".
                                     format(colour('TASK:', ['bright'])),
                                     datefmt="%Y-%m-%d %H:%M:%S")

        def format(self, record):
            if record.levelno == LOG_TASK:
                return self.task_fmt.format(record)
            if record.levelno >= logging.ERROR:
                return self.err_fmt.format(record)
            elif record.levelno >= logging.WARNING:
                return self.warn_fmt.format(record)
            elif record.levelno >= logging.INFO:
                return self.info_fmt.format(record)
            elif record.levelno >= logging.DEBUG:
                return self.debug_fmt.format(record)
            else:
                return self.default_fmt.format(record)

    class ColourlessFormatter(SpecialFormatter):
        """Log file output rules (removes all colour characters)."""
        ansi_escape = re.compile(r'\x1b[^m]*m')
        fmt = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                datefmt="%Y-%m-%d %H:%M:%S")

        default_fmt = logging.Formatter(fmt="[%(asctime)s] INFO: %(message)s",
                                        datefmt="%Y-%m-%d %H:%M:%S")
        debug_fmt = logging.Formatter(fmt="[%(asctime)s] DEBUG: %(message)s",
                                      datefmt="%Y-%m-%d %H:%M:%S")
        info_fmt = logging.Formatter(fmt="[%(asctime)s] INFO: %(message)s",
                                     datefmt="%Y-%m-%d %H:%M:%S")
        warn_fmt = logging.Formatter(fmt="[%(asctime)s] WARNING: %(message)s",
                                     datefmt="%Y-%m-%d %H:%M:%S")
        err_fmt = logging.Formatter(fmt="[%(asctime)s] ERROR: %(message)s",
                                    datefmt="%Y-%m-%d %H:%M:%S")
        task_fmt = logging.Formatter(fmt="[%(asctime)s] TASK: %(message)s",
                                     datefmt="%Y-%m-%d %H:%M:%S")

        def format(self, record):
            record.msg = self.ansi_escape.sub('', record.msg)
            if record.levelno == LOG_TASK:
                return self.task_fmt.format(record)
            if record.levelno >= logging.ERROR:
                return self.err_fmt.format(record)
            elif record.levelno >= logging.WARNING:
                return self.warn_fmt.format(record)
            elif record.levelno >= logging.INFO:
                return self.info_fmt.format(record)
            elif record.levelno >= logging.DEBUG:
                return self.debug_fmt.format(record)
            else:
                return self.default_fmt.format(record)

    # setup general properties of loggers
    timestamp_logger = logging.getLogger('timestamp')
    timestamp_logger.setLevel(logging.DEBUG if debug else logging.INFO)

    no_timestamp_logger = logging.getLogger('no_timestamp')
    no_timestamp_logger.setLevel(logging.DEBUG if debug else logging.INFO)

    warning_logger = logging.getLogger('warnings')
    warning_logger.setLevel(logging.DEBUG if debug else logging.INFO)

    # Pass through tqdm before writing to stdout to fix line return issues.
    tqdm_stream = TqdmStream()

    # setup logging to console
    timestamp_stream_logger = logging.StreamHandler(tqdm_stream)
    timestamp_stream_logger.setFormatter(SpecialFormatter())
    timestamp_logger.addHandler(timestamp_stream_logger)

    no_timestamp_stream_logger = logging.StreamHandler(tqdm_stream)
    no_timestamp_stream_logger.setFormatter(None)
    no_timestamp_logger.addHandler(no_timestamp_stream_logger)

    timestamp_logger.is_silent = False
    no_timestamp_stream_logger.is_silent = False
    if silent:
        timestamp_logger.is_silent = True
        timestamp_stream_logger.setLevel(logging.ERROR)
        no_timestamp_stream_logger.is_silent = True

    if log_dir:
        make_sure_path_exists(log_dir)
        timestamp_file_logger = logging.FileHandler(os.path.join(log_dir,
                                                                 log_file), 'a')
        timestamp_file_logger.setFormatter(ColourlessFormatter())
        timestamp_logger.addHandler(timestamp_file_logger)

        no_timestamp_file_logger = logging.FileHandler(os.path.join(log_dir,
                                                                    log_file), 'a')
        no_timestamp_file_logger.setFormatter(None)
        no_timestamp_logger.addHandler(no_timestamp_file_logger)

        warning_fh = logging.FileHandler(os.path.join(log_dir,
                                                      log_file.replace('.log', '.warnings.log')), 'a')
        warning_fh.setFormatter(ColourlessFormatter())
        warning_logger.addHandler(warning_fh)

    timestamp_logger.info('%s v%s' % (program_name, version))
    base_name = ntpath.basename(sys.argv[0])
    if base_name == '__main__.py':
        prog_name = __name__.split('.')[0]
    else:
        prog_name = base_name
    timestamp_logger.info(f'{prog_name} {" ".join(sys.argv[1:])}')


class TqdmStream(object):
    """Stream handler for tqdm and logging. Fixes line return issues.

    References
    ----------
        https://github.com/tqdm/tqdm/issues/313
    """

    def __init__(self):
        tqdm(disable=True, total=0)  # initialise internal lock

    @classmethod
    def write(cls, msg):
        tqdm.write(msg, end='')

    # is this required?
    # @classmethod
    # def flush(_):
    #   pass



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.config.config import AR122_MARKERS, BAC120_MARKERS, TIGRFAM_HMMS, PFAM_HMM_DIR
from gtdbtk.config.output import PATH_AR122_MARKER_INFO, PATH_BAC120_MARKER_INFO


class MarkerInfoFile(object):
    """Store the GTDB-Tk RED dictionary."""

    marker_paths = {"PFAM": os.path.join(PFAM_HMM_DIR, 'individual_hmms'),
                    "TIGRFAM": os.path.join(os.path.dirname(TIGRFAM_HMMS), 'individual_hmms')}

    def __init__(self, path: str, markers: dict):
        self.path = path
        self.markers = self._parse_markers(markers)

    def _parse_markers(self, markers):
        out = dict()
        for db_marker in sorted(markers):
            for marker in markers[db_marker]:
                marker_id = marker[0:marker.rfind('.')]
                marker_path = os.path.join(self.marker_paths[db_marker], marker)

                # get marker name, description, and size
                with open(marker_path) as fh:
                    for line in fh:
                        if line.startswith("NAME  "):
                            marker_name = line.split("  ")[1].strip()
                        elif line.startswith("DESC  "):
                            marker_desc = line.split("  ")[1].strip()
                        elif line.startswith("LENG  "):
                            marker_size = int(line.split("  ")[1].strip())
                            break

                out[marker_id] = {'name': marker_name,
                                  'desc': marker_desc,
                                  'size': marker_size,
                                  'path': marker_path}
        return out

    def write(self):
        """Write the file to disk, note that domain is omitted."""
        make_sure_path_exists(os.path.dirname(self.path))
        with open(self.path, 'w') as fh:
            fh.write('Marker ID\tName\tDescription\tLength (bp)\n')
            for marker_id, marker_d in sorted(self.markers.items()):
                row = [marker_id, marker_d['name'], marker_d['desc'], str(marker_d['size'])]
                fh.write('\t'.join(row) + '\n')


class MarkerInfoFileAR122(MarkerInfoFile):
    """Marker information for the AR122 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_AR122_MARKER_INFO.format(prefix=prefix))
        super().__init__(path, AR122_MARKERS)


class MarkerInfoFileBAC120(MarkerInfoFile):
    """Marker information for the BAC120 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_BAC120_MARKER_INFO.format(prefix=prefix))
        super().__init__(path, BAC120_MARKERS)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import os
import random
import shutil
import tempfile
import unittest

import numpy as np
from dendropy.simulate import treesim

from gtdbtk import tools
from gtdbtk.tools import TreeTraversal, calculate_patristic_distance


class TestTools(unittest.TestCase):

    def test_add_ncbi_prefix(self):
        refname = 'GCF_123.1'
        self.assertEqual(tools.add_ncbi_prefix(refname), 'RS_GCF_123.1')
        refname = 'GCA_456.1'
        self.assertEqual(tools.add_ncbi_prefix(refname), 'GB_GCA_456.1')
        refname = 'genome_1'
        self.assertEqual(tools.add_ncbi_prefix(refname), refname)

    def test_splitchunks(self):
        test_dict = {'k1': 'v1', 'k2': 'v2',
                     'k3': 'v3', 'k4': 'v4'}
        my_gen = tools.splitchunks(test_dict, 2)
        self.assertEqual(len(next(my_gen)), 2)

    def test_splitchunks_list(self):
        test_list = [1, 2, 3, 4, 5, 6]
        my_gen = tools.splitchunks_list(test_list, 2)
        self.assertEqual(len(next(my_gen)), 3)

    def test_generateTempTableName(self):
        self.assertEqual(len(tools.generateTempTableName()), 24)

    def test_merge_two_dicts(self):
        a = {'k1': 'v1', 'k2': 'v2'}
        b = {'k3': 'v3', 'k4': 'v4'}
        join_dict = tools.merge_two_dicts(a, b)
        self.assertIn('k1', join_dict)
        self.assertIn('k3', join_dict)
        self.assertEqual(len(join_dict), 4)

    def test_sha256(self):
        file = 'gtdbtk/tests/data/genomes/genome_1.fna'
        self.assertEqual(tools.sha256(
            file), '64a8a537d4d964366d7eecefc7cb806f82db5544')

    def test_file_has_checksum(self):
        """Test that the checksum read equals the file contents"""
        dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        try:
            path_file = os.path.join(dir_tmp, 'file.txt')
            with open(path_file, 'w') as f:
                f.write('This is a test!\nFoo\n')
            with open(path_file + '.sha256', 'w') as f:
                f.write('ac4ecc7c839524f886136fbe2e3e17ab6a806a25')
            self.assertTrue(tools.file_has_checksum(path_file, '.sha256'))
        finally:
            shutil.rmtree(dir_tmp)

    def test_file_has_checksum__mismatch(self):
        """Test that the function fails if the checksum mismatches"""
        dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        try:
            path_file = os.path.join(dir_tmp, 'file.txt')
            with open(path_file, 'w') as f:
                f.write('This is a test!\nFoo\n')
            with open(path_file + '.sha256', 'w') as f:
                f.write('asd123')
            self.assertFalse(tools.file_has_checksum(path_file, '.sha256'))
        finally:
            shutil.rmtree(dir_tmp)

    def test_file_has_checksum__file_error(self):
        """Test that the function fails if either file doesn't exist"""
        dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        try:
            path_file = os.path.join(dir_tmp, 'file.txt')
            with open(path_file, 'w') as f:
                f.write('This is a test!\nFoo\n')
            self.assertFalse(tools.file_has_checksum(path_file, '.sha256'))
            self.assertFalse(tools.file_has_checksum('/dev/null/foo', '.sha256'))
        finally:
            shutil.rmtree(dir_tmp)

    def test_get_leaf_nodes(self):
        tree = treesim.birth_death_tree(birth_rate=1.0, death_rate=0.5, num_extant_tips=500)

        all_nodes = list(tree.postorder_node_iter())
        random.shuffle(all_nodes)

        tt = TreeTraversal()
        for node in all_nodes:
            true = frozenset(node.leaf_nodes())
            test = tt.get_leaf_nodes(node)
            self.assertEqual(true, test)

    def test_calculate_patristic_distance(self):
        tree = treesim.birth_death_tree(birth_rate=1.0, death_rate=0.5, num_extant_tips=100)
        for edge in tree.postorder_edge_iter():
            edge.length += np.random.random()
        pdm = tree.phylogenetic_distance_matrix()

        # Generate test cases.
        for int_node in tree.internal_nodes():
            leaf_nodes = int_node.leaf_nodes()
            if len(leaf_nodes) < 2:
                continue

            # Select test data.
            random.shuffle(leaf_nodes)
            qry_node = leaf_nodes[0]
            ref_nodes = leaf_nodes[1:int(np.ceil(len(leaf_nodes) * 0.5))]

            # Calculate the true/test data.
            true = dict()
            for ref_node in ref_nodes:
                true[ref_node] = pdm.patristic_distance(qry_node.taxon,
                                                        ref_node.taxon)
            test = calculate_patristic_distance(qry_node, ref_nodes)

            # Verify that it's correct.
            self.assertSetEqual(set(test.keys()), set(true.keys()))
            for k in test:
                self.assertAlmostEqual(true[k], test[k])


if __name__ == '__main__':
    unittest.main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
from typing import Dict

from gtdbtk.biolib_lite.common import make_sure_path_exists
from gtdbtk.config.config import RED_DIST_ARC_DICT, RED_DIST_BAC_DICT
from gtdbtk.config.output import PATH_AR122_RED_DICT, PATH_BAC120_RED_DICT


class REDDictFile(object):
    """Store the GTDB-Tk RED dictionary."""

    def __init__(self, path: str, red_dict: Dict[str, float]):
        self.path = path
        self.data = red_dict

    def write(self):
        """Write the file to disk, note that domain is omitted."""
        make_sure_path_exists(os.path.dirname(self.path))
        with open(self.path, 'w') as fh:
            fh.write(f'Phylum\t{self.data["p__"]}\n')
            fh.write(f'Class\t{self.data["c__"]}\n')
            fh.write(f'Order\t{self.data["o__"]}\n')
            fh.write(f'Family\t{self.data["f__"]}\n')
            fh.write(f'Genus\t{self.data["g__"]}\n')


class REDDictFileAR122(REDDictFile):
    """Store the RED dictionary for the AR122 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_AR122_RED_DICT.format(prefix=prefix))
        super().__init__(path, RED_DIST_ARC_DICT)


class REDDictFileBAC120(REDDictFile):
    """Store the RED dictionary for the BAC120 marker set."""

    def __init__(self, out_dir: str, prefix: str):
        path = os.path.join(out_dir, PATH_BAC120_RED_DICT.format(prefix=prefix))
        super().__init__(path, RED_DIST_BAC_DICT)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import argparse
import os
import random
import shutil
import string
import tempfile
import unittest

from gtdbtk.biolib_lite.logger import logger_setup
from gtdbtk.config.output import *
from gtdbtk.main import OptionsParser


class TestCli(unittest.TestCase):

    def setUp(self):
        self.identify_dir_reference = os.path.join(os.path.dirname(__file__), 'data/identify_dir_reference/')
        self.align_dir_reference = 'tests/data/align_dir_reference/'
        self.genome_dir = 'gtdbtk/tests/data/genomes/'

        self.options = argparse.ArgumentParser()
        self.options.batchfile = None
        self.options.prefix = 'gtdbtk'
        self.options.cpus = 1
        self.options.extension = 'fna'
        self.options.debug = False
        self.options.force = False
        self.options.genes = False
        self.options.write_single_copy_genes = False

        # align option
        self.options.skip_gtdb_refs = False
        self.options.taxa_filter = None
        self.options.custom_msa_filters = False
        self.options.skip_trimming = False
        self.options.min_consensus = None
        self.options.min_perc_taxa = None
        self.options.skip_gtdb_refs = False
        self.options.cols_per_gene = None
        self.options.max_consensus = None
        self.options.min_perc_aa = 50
        self.options.rnd_seed = 42
        self.options.outgroup_taxon = None

        # classify options
        self.options.scratch_dir = None
        self.options.keep_ref_red = None
        self.options.pplacer_cpus = None
        self.options.min_af = None

        # infer options
        self.options.prot_model = 'WAG'
        self.options.no_support = False
        self.options.no_gamma = True

        self.version = ' unittest'
        self.optionparser = OptionsParser(self.version)
        logger_setup(None, "gtdbtk.log", "GTDB-Tk", self.version, True)
        # self.generic_out_path = 'tests/data/results'
        self.generic_out_path = tempfile.mkdtemp(prefix='gtdbtk_tmp_')

    def tearDown(self):
        shutil.rmtree(self.generic_out_path)


    def test_identify(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))
        identify_options = self.options
        identify_options.genome_dir = self.genome_dir
        identify_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'identify')
        self.optionparser.identify(identify_options)

        ar122_marker_path = os.path.join(self.options.out_dir,
                                         PATH_AR122_MARKER_SUMMARY.format(prefix=self.options.prefix))

        self.assertTrue(os.path.isfile(
            os.path.join(self.options.out_dir, PATH_BAC120_MARKER_SUMMARY.format(prefix=self.options.prefix))))
        self.assertTrue(os.path.isfile(ar122_marker_path))

        results = {}
        with open(ar122_marker_path, 'r') as f:
            f.readline()
            for line in f:
                infos = line.split('\t', 1)
                results[infos[0]] = infos[1]
        self.assertTrue(results.get('genome_1').startswith('120\t2\t0\t'))

    def test_align(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))
        align_options = self.options
        align_options.identify_dir = self.identify_dir_reference
        align_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'align')
        self.optionparser.align(align_options)
        path_user_msa = os.path.join(align_options.out_dir, PATH_AR122_USER_MSA.format(prefix=align_options.prefix))
        self.assertTrue(os.path.isfile(path_user_msa))
        with open(path_user_msa, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertTrue(len(last_line) > 4500)
        self.assertTrue(len(last_line) < 5500)
        self.assertTrue('-' in last_line)
        self.assertFalse(any(char.isdigit() for char in last_line))

    def test_identify_align(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))

        identify_options = self.options
        identify_options.genome_dir = self.genome_dir
        identify_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'identify')
        self.optionparser.identify(identify_options)

        align_options = self.options
        align_options.identify_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'identify')
        align_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'align')
        self.optionparser.align(align_options)
        path_user_msa = os.path.join(align_options.out_dir, PATH_AR122_USER_MSA.format(prefix=align_options.prefix))
        self.assertTrue(os.path.isfile(path_user_msa))
        with open(path_user_msa, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertTrue(len(last_line) > 4500)
        self.assertTrue(len(last_line) < 5500)
        self.assertTrue('-' in last_line)
        self.assertFalse(any(char.isdigit() for char in last_line))

    def test_identify_align_classify(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))

        identify_options = self.options
        identify_options.genome_dir = self.genome_dir
        identify_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'identify')
        self.optionparser.identify(identify_options)

        align_options = self.options
        align_options.identify_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'identify')
        align_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'align')
        self.optionparser.align(align_options)
        path_user_msa = os.path.join(align_options.out_dir, PATH_AR122_USER_MSA.format(prefix=align_options.prefix))
        self.assertTrue(os.path.isfile(path_user_msa))
        with open(path_user_msa, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertTrue(len(last_line) > 4500)
        self.assertTrue(len(last_line) < 5500)
        self.assertTrue('-' in last_line)
        self.assertFalse(any(char.isdigit() for char in last_line))

        classify_options = self.options
        classify_options.genome_dir = self.genome_dir
        classify_options.split_tree = False
        classify_options.align_dir = align_options.out_dir
        classify_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'classify')
        classify_options.recalculate_red = False
        self.optionparser.classify(classify_options)
        summary_out = os.path.join(classify_options.out_dir,
                                   PATH_AR122_SUMMARY_OUT.format(prefix=classify_options.prefix))
        self.assertTrue(summary_out)
        with open(summary_out, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        infos = last_line.split('\t')
        self.assertEqual(len(infos), 20)
        self.assertTrue(infos[1].startswith('d__Archaea'))

    def test_classify_wf(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))
        classify_wf_options = self.options
        classify_wf_options.genome_dir = self.genome_dir
        classify_wf_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'classify_wf')
        self.optionparser.identify(classify_wf_options)
        classify_wf_options.identify_dir = classify_wf_options.out_dir
        classify_wf_options.align_dir = classify_wf_options.out_dir
        classify_wf_options.taxa_filter = None
        classify_wf_options.custom_msa_filters = False
        classify_wf_options.min_consensus = None
        classify_wf_options.min_perc_taxa = None
        classify_wf_options.skip_gtdb_refs = False
        classify_wf_options.cols_per_gene = None
        classify_wf_options.max_consensus = None
        classify_wf_options.recalculate_red = False
        classify_wf_options.split_tree = False
        self.optionparser.align(classify_wf_options)
        self.optionparser.classify(classify_wf_options)
        summary_out = os.path.join(classify_wf_options.out_dir,
                                   PATH_AR122_SUMMARY_OUT.format(prefix=classify_wf_options.prefix))
        self.assertTrue(os.path.isfile(summary_out))
        with open(summary_out, 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        infos = last_line.split('\t')
        self.assertEqual(len(infos), 20)
        self.assertTrue(infos[1].startswith('d__Archaea'))

    def test_infer(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))
        infer_options = self.options
        path_user_msa = PATH_AR122_USER_MSA.format(prefix=self.options.prefix)
        infer_options.msa_file = os.path.join(self.align_dir_reference, path_user_msa)
        infer_options.out_dir = os.path.join(self.generic_out_path, tmp_folder, 'infer')
        infer_options.gamma = False
        # if not os.path.isdir(infer_options.out_dir):
        #     os.makedirs(infer_options.out_dir)
        self.optionparser.infer(infer_options)
        with open(os.path.join(infer_options.out_dir, PATH_TREE_LOG.format(prefix=self.options.prefix)), 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertEqual(last_line.strip(), 'TreeCompleted')
        with open(os.path.join(infer_options.out_dir, PATH_UNROOTED_TREE.format(prefix=self.options.prefix)), 'r') as f:
            lines = f.read().splitlines()
            last_line = lines[-1]
        self.assertTrue('genome_1' in last_line)
        self.assertTrue('genome_2' in last_line)
        self.assertTrue('genome_3' in last_line)

    def test_de_novo_wf(self):
        tmp_folder = ''.join(random.choice(
            string.ascii_uppercase + string.digits) for _ in range(10))
        de_novo_wf_options = self.options
        de_novo_wf_options.genome_dir = self.genome_dir
        de_novo_wf_options.suffix = ".ar122"
        de_novo_wf_options.gamma = False
        de_novo_wf_options.out_dir = os.path.join(
            self.generic_out_path, tmp_folder, 'de_novo_wf')
        de_novo_wf_options.identify_dir = de_novo_wf_options.out_dir
        de_novo_wf_options.msa_file = os.path.join(
            de_novo_wf_options.out_dir, de_novo_wf_options.prefix + de_novo_wf_options.suffix + ".user_msa.fasta")
        self.optionparser.identify(de_novo_wf_options)
        self.optionparser.align(de_novo_wf_options)
        self.optionparser.infer(de_novo_wf_options)

    def test_root(self):
        """Test that rooting is successful when called through the CLI"""
        options = argparse.ArgumentParser()
        options.input_tree = 'tests/data/pplacer_dir_reference/gtdbtk.ar122.classify.tree'
        options.outgroup_taxon = 'p__Altarchaeota'
        options.output_tree = os.path.join(self.generic_out_path, 'test.rooted.tree')
        options.custom_taxonomy_file = None
        options.gtdbtk_classification_file = None
        self.optionparser.root(options)
        self.assertTrue(os.path.isfile(options.output_tree))


if __name__ == '__main__':
    unittest.main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import multiprocessing as mp
import os
import subprocess

from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.io.marker.tophit import TopHitTigrFile
from gtdbtk.tools import sha256, file_has_checksum, tqdm_log


class TigrfamSearch(object):
    """Runs TIGRFAM HMMs over a set of genomes."""

    def __init__(self,
                 threads,
                 tigrfam_hmms,
                 protein_file_suffix,
                 tigrfam_suffix,
                 tigrfam_top_hit_suffix,
                 checksum_suffix,
                 output_dir):
        """Initialization."""
        self.logger = logging.getLogger('timestamp')
        self.warnings = logging.getLogger('warnings')
        self.threads = threads
        self.cpus_per_genome = 1
        self.tigrfam_hmms = tigrfam_hmms
        self.protein_file_suffix = protein_file_suffix
        self.tigrfam_suffix = tigrfam_suffix
        self.tigrfam_top_hit_suffix = tigrfam_top_hit_suffix
        self.checksum_suffix = checksum_suffix
        self.output_dir = output_dir
        self.version = self._get_version()

    def _get_version(self):
        """ get HMMER version."""
        try:
            env = os.environ.copy()
            proc = subprocess.Popen(['hmmsearch', '-h'], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, env=env, encoding='utf-8')

            output, error = proc.communicate()
            for line in output.split('\n'):
                if line.startswith('# HMMER'):
                    version = line.split(';')[0].replace('# HMMER', '').strip()
                    return version
            return "(version unavailable)"
        except:
            return "(version unavailable)"

    def _topHit(self, tigrfam_file):
        """Determine top hits to TIGRFAMs.

        A gene is assigned to a single TIGRFAM
        family. This will be the top hit among
        all TIGRFAM HMMs and pass the threshold
        for the HMM.

        Parameters
        ----------
        tigrfam_file : str
            Name of file containing hits to TIGRFAM HMMs.
        """
        assembly_dir, filename = os.path.split(tigrfam_file)
        genome_id = filename.replace(self.tigrfam_suffix, '')
        tophit_file = TopHitTigrFile(self.output_dir, genome_id)

        # Populate the top-hit file.
        with open(tigrfam_file, 'r') as fh_tigrfam:
            for line in fh_tigrfam:
                if line[0] == '#':
                    continue

                line_split = line.split()
                gene_id = line_split[0]
                hmm_id = line_split[3]
                evalue = float(line_split[4])
                bitscore = float(line_split[5])
                tophit_file.add_hit(gene_id, hmm_id, evalue, bitscore)

        # Write the top-hit file to disk and calculate checksum.
        tophit_file.write()

    def _workerThread(self, queueIn, queueOut, n_skipped):
        """Process each data item in parallel."""
        while True:
            gene_file = queueIn.get(block=True, timeout=None)
            if gene_file is None:
                break

            assembly_dir, filename = os.path.split(gene_file)
            genome_id = filename.replace(self.protein_file_suffix, '')
            genome_dir = os.path.join(self.output_dir, genome_id)
            output_hit_file = os.path.join(genome_dir, filename.replace(self.protein_file_suffix,
                                                                        self.tigrfam_suffix))

            hmmsearch_out = os.path.join(genome_dir, filename.replace(self.protein_file_suffix,
                                                                      '_tigrfam.out'))

            # Check if this has already been processed.
            out_files = (output_hit_file, hmmsearch_out, TopHitTigrFile.get_path(self.output_dir, genome_id))
            if all([file_has_checksum(x) for x in out_files]):
                self.warnings.info(f'Skipped TIGRFAM processing for: {genome_id}')
                with n_skipped.get_lock():
                    n_skipped.value += 1

            else:
                args = ['hmmsearch', '-o', hmmsearch_out, '--tblout', output_hit_file,
                        '--noali', '--notextw', '--cut_nc', '--cpu',
                        str(self.cpus_per_genome), self.tigrfam_hmms, gene_file]
                p = subprocess.Popen(args, stdout=subprocess.PIPE, encoding='utf-8')
                stdout, stderr = p.communicate()

                if p.returncode != 0:
                    raise GTDBTkExit(f'Non-zero exit code returned when running hmsearch: {stdout}')

                # calculate checksum
                for out_file in [output_hit_file, hmmsearch_out]:
                    checksum = sha256(out_file)
                    with open(out_file + self.checksum_suffix, 'w') as fh:
                        fh.write(checksum)

                # identify top hit for each gene
                self._topHit(output_hit_file)

            # allow results to be processed or written to file
            queueOut.put(gene_file)

    def _writerThread(self, numDataItems, writerQueue):
        """Store or write results of worker threads in a single thread."""
        with tqdm_log(total=numDataItems, unit='genome') as p_bar:
            for _ in iter(writerQueue.get, None):
                p_bar.update()

    def run(self, gene_files):
        """Annotate genes with TIGRFAM HMMs.

        Parameters
        ----------
        gene_files : iterable
            Gene files in FASTA format to process.
        """
        if len(gene_files) == 0:
            raise GTDBTkExit('There are no genomes to process.')
        self.cpus_per_genome = max(1, int(self.threads / len(gene_files)))

        # populate worker queue with data to process
        workerQueue = mp.Queue()
        writerQueue = mp.Queue()
        n_skipped = mp.Value('i', 0)

        for f in gene_files:
            workerQueue.put(f)

        for _ in range(self.threads):
            workerQueue.put(None)

        try:
            workerProc = [mp.Process(target=self._workerThread, args=(
                workerQueue, writerQueue, n_skipped)) for _ in range(self.threads)]
            writeProc = mp.Process(target=self._writerThread, args=(
                len(gene_files), writerQueue))

            writeProc.start()

            for p in workerProc:
                p.start()

            for p in workerProc:
                p.join()

            writerQueue.put(None)
            writeProc.join()

            for proc in workerProc:
                if proc.exitcode != 0:
                    raise GTDBTkExit(
                        'An error was encountered while running hmmsearch.')

        except Exception:
            for p in workerProc:
                p.terminate()

            writeProc.terminate()
            raise

        if n_skipped.value > 0:
            genome_s = 'genome' if n_skipped.value == 1 else 'genomes'
            self.logger.warning(f'TIGRFAM skipped {n_skipped.value:,} {genome_s} '
                                f'due to pre-existing data, see warnings.log')



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################
import os
import random
import subprocess
import tempfile
import unittest

from gtdbtk.config.config import CONCAT_AR122, CONCAT_BAC120, MASK_AR122, MASK_DIR, MASK_BAC120
from gtdbtk.io.classify_summary import ClassifySummaryFileAR122
from gtdbtk.tools import sha256


class TestMain(unittest.TestCase):
    """Used to test the CLI. Currently just checks all exit codes."""

    genome_dir = 'gtdbtk/tests/data/genomes/'
    genome_dir_gz = 'tests/data/genomes_gz/'
    cpus = '30'

    def setUp(self):
        """Create a new temporary directory for each method."""
        self.dir_tmp_obj = tempfile.TemporaryDirectory(prefix='gtdbtk_tmp')
        self.dir_tmp = self.dir_tmp_obj.name

    def tearDown(self):
        """Cleanup after each method finishes."""
        self.dir_tmp_obj.cleanup()

    def test_classify_wf(self):
        args = ['python', '-m', 'gtdbtk', 'classify_wf', '--genome_dir',
                self.genome_dir, '--out_dir', self.dir_tmp, '--cpus', self.cpus]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        classify_file = ClassifySummaryFileAR122(out_dir=self.dir_tmp, prefix='gtdbtk')
        classify_file.read()

        expected = {
            'genome_1': 'd__Archaea;p__Methanobacteriota;c__Methanobacteria;o__Methanobacteriales;f__Methanobacteriaceae;g__Methanobrevibacter;s__Methanobrevibacter ruminantium',
            'genome_2': 'd__Archaea;p__Thermoplasmatota;c__Thermoplasmata;o__Methanomassiliicoccales;f__Methanomethylophilaceae;g__VadinCA11;s__VadinCA11 sp002498365',
            'genome_3': 'd__Archaea;p__Thermoplasmatota;c__Thermoplasmata;o__Methanomassiliicoccales;f__Methanomethylophilaceae;g__VadinCA11;s__VadinCA11 sp002498365'
        }
        actual = {k: v.classification for k, v in classify_file.rows.items()}
        self.assertDictEqual(expected, actual)


    def test_de_novo_wf(self):
        args = ['python', '-m', 'gtdbtk', 'de_novo_wf', '--archaea', '--genome_dir',
                self.genome_dir, '--out_dir', self.dir_tmp, '--cpus', self.cpus,
                '--outgroup_taxon', 'p__Altarchaeota']
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_identify(self):
        """Duplicated in classify"""
        pass

    def test_identify_gzipped_genomes(self):
        """Gene calling results should match if genomes are gzipped."""
        args = ['python', '-m', 'gtdbtk', 'identify', '--genome_dir',
                self.genome_dir_gz, '--out_dir', self.dir_tmp, '--cpus', self.cpus,
                '--extension', 'gz']
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

    def test_align(self):
        """Duplicated in classify"""
        pass

    def test_classify(self):
        identify_dir = os.path.join(self.dir_tmp, 'identify')
        args = ['python', '-m', 'gtdbtk', 'identify', '--genome_dir',
                self.genome_dir, '--out_dir', identify_dir, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        align_dir = os.path.join(self.dir_tmp, 'align')
        args = ['python', '-m', 'gtdbtk', 'align', '--identify_dir',
                identify_dir, '--out_dir', align_dir, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        classify_dir = os.path.join(self.dir_tmp, 'classify')
        args = ['python', '-m', 'gtdbtk', 'classify', '--genome_dir',
                self.genome_dir, '--align_dir', align_dir, '--out_dir', classify_dir,
                '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_infer(self):
        """Duplicated in decorate"""
        pass

    def test_root(self):
        """Duplicated in decorate"""
        pass

    def test_decorate(self):
        identify_dir = os.path.join(self.dir_tmp, 'identify')
        args = ['python', '-m', 'gtdbtk', 'identify', '--genome_dir',
                self.genome_dir, '--out_dir', identify_dir, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        align_dir = os.path.join(self.dir_tmp, 'align')
        args = ['python', '-m', 'gtdbtk', 'align', '--identify_dir',
                identify_dir, '--out_dir', align_dir, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        msa_file = os.path.join(align_dir, 'align', 'gtdbtk.ar122.msa.fasta')
        infer_dir = os.path.join(self.dir_tmp, 'infer')
        args = ['python', '-m', 'gtdbtk', 'infer', '--msa_file',
                msa_file, '--out_dir', infer_dir, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        input_tree = os.path.join(infer_dir, 'infer', 'intermediate_results', 'gtdbtk.unrooted.tree')
        rooted_tree = os.path.join(self.dir_tmp, 'root.tree')
        args = ['python', '-m', 'gtdbtk', 'root', '--input_tree',
                input_tree, '--outgroup_taxon', 'p__Altarchaeota',
                '--output_tree', rooted_tree]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        decorated_tree = os.path.join(self.dir_tmp, 'decorated.tree')
        args = ['python', '-m', 'gtdbtk', 'decorate', '--input_tree',
                rooted_tree, '--output_tree', decorated_tree]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

        # Test infer ranks
        path_infer_tree = os.path.join(self.dir_tmp, 'infer.tree')
        args = ['python', '-m', 'gtdbtk', 'infer_ranks', '--input_tree', decorated_tree,
                '--ingroup_taxon', 'p__Altarchaeota', '--output_tree', path_infer_tree]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_infer_ranks(self):
        """Duplicated in decorate."""
        pass

    def test_ani_rep(self):
        args = ['python', '-m', 'gtdbtk', 'ani_rep', '--genome_dir', self.genome_dir,
                '--out_dir', self.dir_tmp, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_trim_msa(self):
        path_out = os.path.join(self.dir_tmp, 'msa.faa')
        args = ['python', '-m', 'gtdbtk', 'trim_msa', '--untrimmed_msa', CONCAT_AR122,
                '--output', path_out, '--reference_mask', 'arc']
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_trim_msa_reference_mask_arc(self):
        with open(os.path.join(MASK_DIR, MASK_AR122), 'r') as f:
            mask = [x == '1' for x in f.read().strip()]

        path_untrimmed_msa = os.path.join(self.dir_tmp, 'untrimmed_msa.fasta')
        path_output = os.path.join(self.dir_tmp, 'trimmed_msa.fasta')

        choices = 'ARNDCQEGHILKMFPSTWYV-'
        expected = list()
        with open(path_untrimmed_msa, 'w') as f:
            f.write('>test123\n')
            for cur_char in mask:
                choice = random.choice(choices)
                if cur_char:
                    expected.append(choice)
                f.write(choice)

        args = ['python', '-m', 'gtdbtk', 'trim_msa', '--untrimmed_msa', path_untrimmed_msa,
                '--reference_mask', 'arc', '--output', path_output]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        results = dict()
        with open(path_output, 'r') as f:
            import re
            re_hits = re.findall(r'>(.+)\n(.+)\n', f.read())
            for gid, seq in re_hits:
                results[gid] = seq

        expected = {'test123': ''.join(expected)}

        self.assertDictEqual(results, expected)

    def test_trim_msa_reference_mask_bac(self):
        with open(os.path.join(MASK_DIR, MASK_BAC120), 'r') as f:
            mask = [x == '1' for x in f.read().strip()]

        path_untrimmed_msa = os.path.join(self.dir_tmp, 'untrimmed_msa.fasta')
        path_output = os.path.join(self.dir_tmp, 'trimmed_msa.fasta')

        choices = 'ARNDCQEGHILKMFPSTWYV-'
        expected = list()
        with open(path_untrimmed_msa, 'w') as f:
            f.write('>test123\n')
            for cur_char in mask:
                choice = random.choice(choices)
                if cur_char:
                    expected.append(choice)
                f.write(choice)

        args = ['python', '-m', 'gtdbtk', 'trim_msa', '--untrimmed_msa', path_untrimmed_msa,
                '--reference_mask', 'bac', '--output', path_output]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        results = dict()
        with open(path_output, 'r') as f:
            import re
            re_hits = re.findall(r'>(.+)\n(.+)\n', f.read())
            for gid, seq in re_hits:
                results[gid] = seq

        expected = {'test123': ''.join(expected)}

        self.assertDictEqual(results, expected)

    def test_trim_msa_mask_file(self):
        path_untrimmed_msa = os.path.join(self.dir_tmp, 'untrimmed_msa.fasta')
        path_mask_file = os.path.join(self.dir_tmp, 'mask_file.txt')
        path_output = os.path.join(self.dir_tmp, 'trimmed_msa.fasta')

        with open(path_untrimmed_msa, 'w') as f:
            f.write('>genome_1\n')
            f.write('ALGPVW\n')
            f.write('>genome_2\n')
            f.write('WVPGLA\n')

        with open(path_mask_file, 'w') as f:
            f.write('010010\n')

        args = ['python', '-m', 'gtdbtk', 'trim_msa', '--untrimmed_msa', path_untrimmed_msa,
                '--mask_file', path_mask_file, '--output', path_output]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        results = dict()
        with open(path_output, 'r') as f:
            import re
            re_hits = re.findall(r'>(.+)\n(.+)\n', f.read())
            for gid, seq in re_hits:
                results[gid] = seq

        expected = {'genome_1': 'LV', 'genome_2': 'VL'}

        self.assertDictEqual(results, expected)

    def test_export_msa_arc(self):
        """Test that the MSA can be exported when using the CLI."""
        path_output = os.path.join(self.dir_tmp, 'msa.faa')
        args = ['python', '-m', 'gtdbtk', 'export_msa', '--domain', 'arc', '--output', path_output]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        test_hash = sha256(path_output)
        true_hash = sha256(CONCAT_AR122)
        self.assertEqual(test_hash, true_hash)

    def test_export_msa_bac(self):
        """Test that the MSA can be exported when using the CLI."""
        path_output = os.path.join(self.dir_tmp, 'msa.faa')
        args = ['python', '-m', 'gtdbtk', 'export_msa', '--domain', 'bac', '--output', path_output]
        p = subprocess.Popen(args)
        p.wait()
        self.assertEqual(p.returncode, 0)

        test_hash = sha256(path_output)
        true_hash = sha256(CONCAT_BAC120)
        self.assertEqual(test_hash, true_hash)

    def test_test(self):
        args = ['python', '-m', 'gtdbtk', 'test', '--out_dir', self.dir_tmp, '--cpus', self.cpus]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)

    def test_check_install(self):
        args = ['python', '-m', 'gtdbtk', 'check_install']
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        stdout, stderr = proc.communicate()
        self.assertEqual(proc.returncode, 0)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import shutil
import tempfile
import unittest

from gtdbtk.trim_msa import TrimMSA


class TestTrimMSA(unittest.TestCase):

    def setUp(self):
        self.tmp_out_dir = tempfile.mkdtemp()
        self.trim_msa = TrimMSA(cols_per_gene=0.42,
                                min_consensus=0.25,
                                max_consensus=0.95,
                                min_perc_taxa=0.50,
                                rnd_seed=42,
                                min_perc_aa=0.10,
                                out_dir=self.tmp_out_dir)

    def tearDown(self):
        shutil.rmtree(self.tmp_out_dir)

    def test_identify_valid_columns(self):
        """ Test that the expected columns are identified as valid. """
        test_seqs = {'genome_1': 'AAVWAWGG',
                     'genome_2': 'AAVW-WGA',
                     'genome_3': '--AW-GGG'}
        result = self.trim_msa.identify_valid_columns(0, 7, test_seqs)
        self.assertSetEqual(result, {2, 5})



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import shutil
import tempfile
import unittest

from gtdbtk.config.output import PATH_BAC120_MARKER_SUMMARY, PATH_AR122_MARKER_SUMMARY
from gtdbtk.io.marker.copy_number import CopyNumberFile, CopyNumberFileAR122, CopyNumberFileBAC120
from gtdbtk.io.marker.tophit import TopHitPfamFile, TopHitTigrFile, Hit


class TestCopyNumberFile(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')

        # Build a test set of markers
        self.test_markers = {"PFAM": ["PFAM_1.hmm", "PFAM_2.hmm", "PFAM_3.hmm"],
                             "TIGRFAM": ["TIGR_1.HMM", "TIGR_2.HMM"]}

        # Build the copy number file
        self.cn_path = os.path.join(self.dir_tmp, 'cn.tsv')
        self.cn = CopyNumberFile(self.cn_path, 'test5', self.test_markers)

        # Store the hit files
        self.pfam_th = dict()
        self.tigr_th = dict()
        self.faa = dict()

        # Test case for genome_1
        """
        Single Copy: PFAM_1, TIGR_1
        Multi-Unique: PFAM_2
        Multi-Copy: None
        Missing: PFAM_3, TIGR_2
        """
        self.pfam_th['genome_1'] = TopHitPfamFile(os.path.join(self.dir_tmp), 'genome_1')
        self.pfam_th['genome_1'].add_hit('gene_a', 'PFAM_1', 0.05, 100)
        self.pfam_th['genome_1'].add_hit('gene_b', 'PFAM_2', 0.05, 200)
        self.pfam_th['genome_1'].add_hit('gene_c', 'PFAM_2', 0.05, 100)
        self.pfam_th['genome_1'].write()

        self.tigr_th['genome_1'] = TopHitTigrFile(os.path.join(self.dir_tmp), 'genome_1')
        self.tigr_th['genome_1'].add_hit('gene_x', 'TIGR_1', 0.05, 100)

        self.faa['genome_1'] = os.path.join(self.dir_tmp, 'genome_1.faa')
        with open(self.faa['genome_1'], 'w') as fh:
            fh.write('>gene_a\n')
            fh.write('VVVVVV\n')
            fh.write('>gene_b\n')
            fh.write('AAVVPP\n')
            fh.write('>gene_c\n')
            fh.write('AAVVPP\n')
            fh.write('>gene_x\n')
            fh.write('AAAAAA\n')

        # Test case for genome_2
        """
        Single Copy: PFAM_2
        Multi-Unique: TIGR_1, TIGR_2
        Multi-Copy: PFAM_1, PFAM_3
        Missing: None
        """
        self.pfam_th['genome_2'] = TopHitPfamFile(os.path.join(self.dir_tmp), 'genome_2')
        self.pfam_th['genome_2'].add_hit('gene_a', 'PFAM_2', 0.05, 100)
        self.pfam_th['genome_2'].add_hit('gene_w', 'PFAM_1', 0.05, 100)
        self.pfam_th['genome_2'].add_hit('gene_z', 'PFAM_1', 0.05, 100)
        self.pfam_th['genome_2'].add_hit('gene_y', 'PFAM_3', 0.05, 100)
        self.pfam_th['genome_2'].add_hit('gene_x', 'PFAM_3', 0.05, 100)
        self.pfam_th['genome_2'].write()

        self.tigr_th['genome_2'] = TopHitTigrFile(os.path.join(self.dir_tmp), 'genome_2')
        self.tigr_th['genome_2'].add_hit('gene_w', 'TIGR_1', 0.05, 100)
        self.tigr_th['genome_2'].add_hit('gene_x', 'TIGR_1', 0.05, 200)
        self.tigr_th['genome_2'].add_hit('gene_y', 'TIGR_2', 0.01, 100)
        self.tigr_th['genome_2'].add_hit('gene_z', 'TIGR_2', 0.05, 100)

        self.faa['genome_2'] = os.path.join(self.dir_tmp, 'genome_2.faa')
        with open(self.faa['genome_2'], 'w') as fh:
            fh.write('>gene_a\n')
            fh.write('VVVVVV\n')
            fh.write('>gene_w\n')
            fh.write('AAAVVV\n')
            fh.write('>gene_x\n')
            fh.write('AAAVVV\n')
            fh.write('>gene_y\n')
            fh.write('VVVAAA\n')
            fh.write('>gene_z\n')
            fh.write('VVVAAA\n')

        # Add the genomes
        for gid in self.faa:
            self.cn.add_genome(gid, self.faa[gid], pfam_th=self.pfam_th[gid], tigr_th=self.tigr_th[gid])

    def tearDown(self):
        self.cn = None
        self.pfam_th = None
        self.tigr_th = None
        shutil.rmtree(self.dir_tmp)

    def test_add_genome(self):
        genome_1_truth = {'unq': {'PFAM_1': {'hit': Hit('gene_a', 'PFAM_1', 0.05, 100), 'seq': 'VVVVVV'},
                                  'TIGR_1': {'hit': Hit('gene_x', 'TIGR_1', 0.05, 100), 'seq': 'AAAAAA'}
                                  },
                          'muq': {'PFAM_2': {'hit': Hit('gene_b', 'PFAM_2', 0.05, 200), 'seq': 'AAVVPP'}},
                          'mul': {},
                          'mis': {'PFAM_3': None, 'TIGR_2': None}
                          }
        genome_2_truth = {'unq': {'PFAM_2': {'hit': Hit('gene_a', 'PFAM_2', 0.05, 100), 'seq': 'VVVVVV'}},
                          'muq': {'TIGR_1': {'hit': Hit('gene_x', 'TIGR_1', 0.05, 200), 'seq': 'AAAVVV'},
                                  'TIGR_2': {'hit': Hit('gene_y', 'TIGR_2', 0.01, 100), 'seq': 'VVVAAA'}},
                          'mul': {'PFAM_1': None, 'PFAM_3': None},
                          'mis': {}}

        self.assertDictEqual(genome_1_truth, self.cn.genomes['genome_1'])
        self.assertDictEqual(genome_2_truth, self.cn.genomes['genome_2'])

    def test__extract_marker_names(self):
        true = {'PFAM_1', 'PFAM_2', 'PFAM_3', 'TIGR_1', 'TIGR_2'}
        self.assertSetEqual(true, CopyNumberFile._extract_marker_names(self.test_markers))

    def test__merge_hit_files(self):
        pfam_th = TopHitPfamFile(os.path.join(self.dir_tmp), 'genome_1')
        pfam_th.add_hit('gene_a', 'PFAM_1', 0.05, 100)
        pfam_th.add_hit('gene_b', 'PFAM_2', 0.05, 200)
        pfam_th.add_hit('gene_c', 'PFAM_2', 0.05, 100)

        tigr_th = TopHitTigrFile(os.path.join(self.dir_tmp), 'genome_1')
        tigr_th.add_hit('gene_x', 'TIGR_1', 0.05, 100)

        expected = {'TIGR_1': [Hit('gene_x', 'TIGR_1', 0.05, 100)],
                    'PFAM_1': [Hit('gene_a', 'PFAM_1', 0.05, 100)],
                    'PFAM_2': [Hit('gene_b', 'PFAM_2', 0.05, 200),
                               Hit('gene_c', 'PFAM_2', 0.05, 100)]}
        self.assertDictEqual(expected, CopyNumberFile._merge_hit_files(pfam_th, tigr_th))

    def test_get_single_copy_hits(self):
        g1_expected = {'PFAM_1': {'hit': Hit('gene_a', 'PFAM_1', 0.05, 100), 'seq': 'VVVVVV'},
                       'TIGR_1': {'hit': Hit('gene_x', 'TIGR_1', 0.05, 100), 'seq': 'AAAAAA'},
                       'PFAM_2': {'hit': Hit('gene_b', 'PFAM_2', 0.05, 200), 'seq': 'AAVVPP'}}
        g2_expected = {'PFAM_2': {'hit': Hit('gene_a', 'PFAM_2', 0.05, 100), 'seq': 'VVVVVV'},
                       'TIGR_1': {'hit': Hit('gene_x', 'TIGR_1', 0.05, 200), 'seq': 'AAAVVV'},
                       'TIGR_2': {'hit': Hit('gene_y', 'TIGR_2', 0.01, 100), 'seq': 'VVVAAA'}}

        self.assertDictEqual(g1_expected, self.cn.get_single_copy_hits('genome_1'))
        self.assertDictEqual(g2_expected, self.cn.get_single_copy_hits('genome_2'))

    def test_write(self):
        self.cn.write()

        with open(self.cn.path) as fh:
            self.assertEqual('name\tnumber_unique_genes\tnumber_multiple_genes\t'
                             'number_multiple_unique_genes\tnumber_missing_genes\t'
                             'list_unique_genes\tlist_multiple_genes\t'
                             'list_multiple_unique_genes\tlist_missing_genes\n', fh.readline())
            self.assertEqual('genome_1\t2\t0\t1\t2\t'
                             'PFAM_1,TIGR_1\t\tPFAM_2\tPFAM_3,TIGR_2\n', fh.readline())
            self.assertEqual('genome_2\t1\t2\t2\t0\t'
                             'PFAM_2\tPFAM_1,PFAM_3\tTIGR_1,TIGR_2\t\n', fh.readline())

    def test_read(self):
        self.cn.write()

        test_cn = CopyNumberFile(self.cn.path, 'test5', self.test_markers)
        test_cn.read()

        genome_1_truth = {'unq': {'PFAM_1': None, 'TIGR_1': None
                                  },
                          'muq': {'PFAM_2': None},
                          'mul': {},
                          'mis': {'PFAM_3': None, 'TIGR_2': None}
                          }
        genome_2_truth = {'unq': {'PFAM_2': None},
                          'muq': {'TIGR_1': None, 'TIGR_2': None},
                          'mul': {'PFAM_1': None, 'PFAM_3': None},
                          'mis': {}}

        self.assertDictEqual(genome_1_truth, test_cn.genomes['genome_1'])
        self.assertDictEqual(genome_2_truth, test_cn.genomes['genome_2'])


class TestCopyNumberFileAR122(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.cn = CopyNumberFileAR122(self.dir_tmp, 'tst')

    def tearDown(self):
        self.cn = None
        shutil.rmtree(self.dir_tmp)

    def test___init__(self):
        path = os.path.join(self.dir_tmp, PATH_AR122_MARKER_SUMMARY.format(prefix='tst'))
        self.assertEqual(path, self.cn.path)
        self.assertEqual('ar122', self.cn.marker_set)
        self.assertDictEqual({}, self.cn.genomes)
        self.assertEqual(122, len(self.cn.marker_names))


class TestCopyNumberFileBAC120(unittest.TestCase):

    def setUp(self):
        self.dir_tmp = tempfile.mkdtemp(prefix='gtdbtk_tmp_')
        self.cn = CopyNumberFileBAC120(self.dir_tmp, 'tst')

    def tearDown(self):
        self.cn = None
        shutil.rmtree(self.dir_tmp)

    def test___init__(self):
        path = os.path.join(self.dir_tmp, PATH_BAC120_MARKER_SUMMARY.format(prefix='tst'))
        self.assertEqual(path, self.cn.path)
        self.assertEqual('bac120', self.cn.marker_set)
        self.assertDictEqual({}, self.cn.genomes)
        self.assertEqual(120, len(self.cn.marker_names))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import multiprocessing as mp
import os
import queue
import re
import subprocess

from tqdm import tqdm

from gtdbtk.exceptions import PplacerException, TogException
from gtdbtk.tools import get_proc_memory_gb


class Pplacer(object):
    """Phylogenetic placement of genomes into a reference tree
    (http://matsen.fredhutch.org/pplacer/).
    """

    def __init__(self):
        """ Instantiate the class. """
        self.logger = logging.getLogger('timestamp')
        self.version = self._get_version()

    def _get_version(self):
        try:
            env = os.environ.copy()
            proc = subprocess.Popen(['pplacer', '--version'], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, env=env, encoding='utf-8')

            output, error = proc.communicate()
            return output.strip()
        except:
            return "(version unavailable)"

    def run(self, cpus, model, ref_pkg, json_out, msa_file, pplacer_out,
            mmap_file=None):
        """Place genomes into a reference tree.

        Args:
            cpus (int): The number of threads to use.
            model (str): The model to use. PROT: LG, WAG, JTT. NT: GTR.
            ref_pkg (str): The path to the reference package.
            json_out (str): The path to write the json output to.
            msa_file (str): The path to the input MSA file.
            pplacer_out (str): Where to write the pplacer output file.
            mmap_file (str, optional): The path to write a scratch file to.

        Raises:
            PplacerException: if a non-zero exit code, or if the json output
                              file isn't generated.

        """
        args = ['pplacer', '-m', model, '-j', str(cpus), '-c', ref_pkg, '-o',
                json_out, msa_file]
        if mmap_file:
            args.append('--mmap-file')
            args.append(mmap_file)
        self.logger.debug(' '.join(args))

        out_q = mp.Queue()
        pid = mp.Value('i', 0)
        p_worker = mp.Process(target=self._worker, args=(
            args, out_q, pplacer_out, pid))
        p_writer = mp.Process(target=self._writer, args=(out_q, pid))

        try:
            p_worker.start()
            p_writer.start()

            p_worker.join()
            out_q.put(None)
            p_writer.join()

            if p_worker.exitcode != 0:
                raise PplacerException(
                    'An error was encountered while running pplacer.')
        except Exception:
            p_worker.terminate()
            p_writer.terminate()
            raise
        finally:
            if mmap_file:
                os.remove(mmap_file)

        if not os.path.isfile(json_out):
            self.logger.error('pplacer returned a zero exit code but no output '
                              'file was generated.')
            raise PplacerException

    def _worker(self, args, out_q, pplacer_out, pid):
        """The worker thread writes the piped output of pplacer to disk and
        shares it with the writer thread for logging."""
        with subprocess.Popen(args, stdout=subprocess.PIPE, encoding='utf-8') as proc:
            with pid.get_lock():
                pid.value = proc.pid

            with open(pplacer_out, 'w') as fh:
                while True:
                    line = proc.stdout.readline()
                    if not line:
                        break
                    fh.write(f'{line}')
                    out_q.put(line)
            proc.wait()

            if proc.returncode != 0:
                raise PplacerException('An error was encountered while '
                                       'running pplacer, check the log '
                                       'file: {}'.format(pplacer_out))

    def _writer(self, out_q, pid):
        """The writer subprocess is able to report on newly piped events from
        subprocess in the worker thread, and report on memory usage while
        waiting for new commands."""
        states = ['Reading user alignment',
                  'Reading reference alignment',
                  'Pre-masking sequences',
                  'Determining figs',
                  'Allocating memory for internal nodes',
                  'Caching likelihood information on reference tree',
                  'Pulling exponents',
                  'Preparing the edges for baseball',
                  'Placing genomes']
        cur_state = None
        n_total, n_placed = None, 0
        # Create a progress bar which tracks the initialization step of
        # pplacer.
        bar_fmt = '==> {desc}.'
        with tqdm(bar_format=bar_fmt, miniters=1, leave=False) as p_bar:
            p_bar.set_description_str(desc='Step 1 of 9: Starting pplacer')
            while True:
                try:
                    state = out_q.get(block=True, timeout=5)
                    if not state:
                        break
                    elif state.startswith('Running pplacer'):
                        cur_state = 0
                    elif state.startswith("Didn't find any reference"):
                        cur_state = 1
                    elif state.startswith('Pre-masking sequences'):
                        cur_state = 2
                    elif state.startswith('Determining figs'):
                        cur_state = 3
                    elif state.startswith('Allocating memory for internal'):
                        cur_state = 4
                    elif state.startswith('Caching likelihood information'):
                        cur_state = 5
                    elif state.startswith('Pulling exponents'):
                        cur_state = 6
                    elif state.startswith('Preparing the edges'):
                        cur_state = 7
                    elif state.startswith('working on '):
                        cur_state = 8
                    else:
                        cur_state = None

                    # Error, display the log file line.
                    if not cur_state:
                        p_bar.set_description_str(desc=state.strip())

                    # New line written for each placed genome, update the
                    # count.
                    elif cur_state == 8:
                        if not n_total:
                            n_total = int(
                                re.search(r'\((\d+)\/(\d+)\)', state).group(2))
                        n_placed += 1
                        p_bar.set_description_str(desc=f'Step 9 of {len(states)}: placing '
                                                  f'genome {n_placed:,} of {n_total:,} '
                                                  f'({n_placed / n_total:.2%})')

                    # Display the output for the current state.
                    else:
                        p_bar.set_description_str(desc=f'Step {cur_state + 1} of {len(states)}: '
                                                  f'{states[cur_state + 1]}')

                # Report the memory usage if at a memory-reportable state.
                except queue.Empty:
                    if cur_state == 3:
                        virt, res = get_proc_memory_gb(pid.value)
                        if virt and res:
                            p_bar.set_description_str(desc=f'Step {cur_state + 1} of {len(states)} '
                                                      f'{states[4]} ({virt:.2f} GB)')
                        else:
                            p_bar.set_description_str(desc=f'Step {cur_state + 1} of {len(states)} '
                                                           f'{states[4]}')
                    elif cur_state == 4:
                        virt, res = get_proc_memory_gb(pid.value)
                        if virt and res:
                            p_bar.set_description_str(desc=f'Step {cur_state + 1} of {len(states)}: '
                                                      f'{states[5]} ({res:.2f}/{virt:.2f} GB, '
                                                      f'{res / virt:.2%})')
                        else:
                            p_bar.set_description_str(desc=f'Step {cur_state + 1} of {len(states)}: '
                                                           f'{states[5]}')
                except Exception:
                    pass

    def tog(self, pplacer_json_out, tree_file):
        """ Convert the pplacer json output into a newick tree.

        Args:
            pplacer_json_out (str): The path to the output of pplacer.
            tree_file (str): The path to output the newick file to.

        Raises:
            TogException: If a non-zero exit code is returned, or the tree file
                          isn't output.
        """

        args = ['guppy', 'tog', '-o', tree_file, pplacer_json_out]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        proc_out, proc_err = proc.communicate()

        if proc.returncode != 0:
            self.logger.error('An error was encountered while running tog.')
            raise TogException(proc_err)

        if not os.path.isfile(pplacer_json_out):
            self.logger.error('tog returned a zero exit code but no output '
                              'file was generated.')
            raise TogException



import logging
import multiprocessing as mp
import os
import subprocess
import tempfile
from collections import defaultdict

from gtdbtk.config.config import LOG_TASK
from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.external.hmm_aligner import HmmAligner
from gtdbtk.io.marker.copy_number import CopyNumberFile
from gtdbtk.io.marker.tophit import TopHitPfamFile, TopHitTigrFile
from gtdbtk.io.marker_info import MarkerInfoFile
from gtdbtk.tools import tqdm_log


def get_single_copy_hits_worker(job):
    """For a given genome, obtain the PFAM and TIGRFAM tophit files. Use
    this information to determine what hits are single copy.

    Parameters
    ----------
    job : Tuple[str, str, CopyNumberFile]
        The genome id, path to called genes, and domain-specific copy number file object.

    Returns
    -------
    Dict[str, Dict[str, str]]
        dict[marker id][genome id] = sequence
    """
    gid, aa_path, copy_number_file = job

    # Load the marker top hit files.
    marker_genes_dir = os.path.dirname(os.path.dirname(aa_path))
    pfam_tophit_file = TopHitPfamFile(marker_genes_dir, gid)
    tigr_tophit_file = TopHitTigrFile(marker_genes_dir, gid)
    pfam_tophit_file.read()
    tigr_tophit_file.read()

    # Process each of the genes to determine if they are single copy.
    cnf = copy_number_file('/dev/null', None)
    cnf.add_genome(gid, aa_path, pfam_tophit_file, tigr_tophit_file)
    single_copy = cnf.get_single_copy_hits(gid)

    # Store the output
    out = defaultdict(dict)
    for marker_id, marker_d in single_copy.items():
        out[marker_id][gid] = marker_d['seq']
    return out


def get_single_copy_hits(gid_dict: dict, copy_number_file, cpus):
    """Collect all of the single copy hits (both domains) for each genome.

    Parameters
    ----------
    gid_dict : dict
        A dictionary containing information about the genome, indexed by the id.
    copy_number_file : CopyNumberFile
        A domain-specific subclass of the copy number file.
    cpus : int
        The number of CPUs to use in sub-processes.

    Returns
    -------
    Dict[str, Dict[str, str]]
        dict[marker id][genome id] = sequence
    """

    # Generate a queue job jobs.
    queue = list()
    for gid, gid_info in gid_dict.items():
        queue.append((gid,
                      gid_info['aa_gene_path'],
                      copy_number_file))

    # Process the queue.
    with mp.get_context('spawn').Pool(processes=cpus) as pool:
        results = list(tqdm_log(pool.imap_unordered(get_single_copy_hits_worker, queue),
                                total=len(queue), unit='genome'))

    # Re-format the results.
    out = defaultdict(dict)
    for result in results:
        for marker_id, marker_d in result.items():
            for gid, seq in marker_d.items():
                out[marker_id][gid] = seq
    return out


def read_hmmalign_output(output, expected_gids):
    """Reads the output of hmmalign and extracts the alignment.

    Parameters
    ----------
    output : str
        The string output from hmmalign.
    expected_gids : frozenset
        A set containing all of the gids expected in the output.

    Returns
    -------
    Dict[str, str]
        dict[genome id] = sequence
    """
    # In case gids have spaces.
    exp_mapping = {x.split(' ', 1)[0]: x for x in expected_gids}

    # Get the sequences and the mask.
    unmasked = dict()
    mask = None
    for line in output.splitlines():
        splitline = line.split(' ', 1)

        # Sequence
        if splitline[0] in exp_mapping:
            rsplitline = line.rsplit(" ", 1)
            hit_seq = rsplitline[-1]
            unmasked[exp_mapping[splitline[0]]] = hit_seq

        # Mask
        elif line[0:len("#=GC RF")] == "#=GC RF":
            mask = [x == 'x' for x in line.rsplit(' ', 1)[-1]]

    # Sanity check.
    if mask is None:
        raise GTDBTkExit(f'Unable to get mask from hmmalign result file: {output}')
    if len(unmasked) != len(expected_gids):
        raise GTDBTkExit(f'Not all genomes could be aligned: {output}')

    # Mask each of the sequences and return them.
    out = dict()
    for gid, seq in unmasked.items():
        out[gid] = ''.join([s for s, m in zip(seq, mask) if m is True])
    return out


def run_hmm_align_worker(job):
    """A worker process for running hmmalign.

    Parameters
    ----------
    job : Tuple[str, str, str, frozenset]
        The marker id, hmm marker path, called genes path, expected gids.

    Returns
    -------
    List[Tuple[str, str, str]]
        A list containing the (genome id, marker id, sequence).
    """
    marker_id, marker_path, marker_fa, expected_gids = job

    # Run the process and capture stdout.
    args = ["hmmalign", "--outformat", "Pfam", marker_path, marker_fa]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
    stdout, stderr = proc.communicate()

    # Exit if an error was raised.
    if proc.returncode != 0:
        arg_str = ' '.join(args)
        raise GTDBTkExit(f'hmmalign returned a non-zero exit code: {arg_str}')

    # Process the output and return the sequences.
    seqs = read_hmmalign_output(stdout, expected_gids)
    return [(gid, marker_id, seq) for gid, seq in seqs.items()]


def create_concat_alignment(list_seqs, marker_info_file):
    """Create a concatenated alignment where missing markers are filled with gaps.

    Parameters
    ----------
    list_seqs : list[tuple[str, str, str]]
        A list containing the (genome id, marker id, sequence).
    marker_info_file : MarkerInfoFile
        A domain specific subclass of the marker info file.

    Returns
    -------
    Dict[str, str]
        dict[gid] = sequence
    """
    # Index the results.
    d_gid_marker = defaultdict(dict)
    for cur_result in list_seqs:
        for gid, marker_id, seq in cur_result:
            d_gid_marker[gid][marker_id] = seq

    # Create the alignment.
    out = defaultdict(list)
    for gid, cur_marker_d in d_gid_marker.items():
        for marker_id, marker_info in sorted(marker_info_file.markers.items()):
            out[gid].append(cur_marker_d.get(marker_id, '-' * marker_info['size']))
    return {k: ''.join(v) for k, v in out.items()}


def align_marker_set(gid_dict, marker_info_file: MarkerInfoFile, copy_number_file: CopyNumberFile, cpus):
    """Aligns the set of genomes for a specific domain.

    Parameters
    ----------
    gid_dict : dict
        A dictionary containing information about the genome, indexed by the id.
    marker_info_file : MarkerInfoFile
        A domain specific subclass of the marker info file.
    copy_number_file : CopyNumberFile
        A domain-specific subclass of the copy number file.
    cpus : int
        The maximum number of CPUs to use in subprocesses.

    Returns
    -------
    Dict[str, str]
        dict[gid] = sequence
    """
    logger = logging.getLogger('timestamp')

    logger.log(LOG_TASK, f'Generating concatenated alignment for each marker.')
    single_copy_hits = get_single_copy_hits(gid_dict, copy_number_file, cpus)

    with tempfile.TemporaryDirectory(prefix='gtdbtk_tmp_') as dir_tmp:
        # Write each of the markers to disk.
        marker_paths = dict()
        for marker_id, marker_d in single_copy_hits.items():
            cur_path = os.path.join(dir_tmp, f'{marker_id}.fa')
            marker_paths[marker_id] = cur_path
            with open(cur_path, 'w') as fh:
                for cur_gid, cur_seq in marker_d.items():
                    fh.write(f'>{cur_gid}\n{cur_seq}\n')

        # Run hmmalign on all of the markers (in order of largest)
        hmmer_v = HmmAligner.get_version()
        logger.log(LOG_TASK, f'Aligning {len(marker_paths)} identified markers using hmmalign {hmmer_v}.')
        queue = list()
        for marker_id, marker_path in sorted(marker_paths.items(),
                                             key=lambda z: -marker_info_file.markers[z[0]]['size']):
            queue.append((marker_id,
                          marker_info_file.markers[marker_id]['path'],
                          marker_path,
                          frozenset(single_copy_hits[marker_id])))
        with mp.get_context('spawn').Pool(processes=cpus) as pool:
            results = list(tqdm_log(pool.imap_unordered(run_hmm_align_worker, queue),
                                    total=len(queue), unit='marker'))

    # Create the concatenated alignment.
    return create_concat_alignment(results, marker_info_file)
